<?php
/**
    @file
    @brief Contient des classes permettant l'utilisation de Curl en orienté objet
    @details
		La classe HttpClient encapsule une resource curl et permet d'utiliser celui-ci
		en orienté objet.
		2 types d'exceptions, toutes 2 déclarées ici, peuvent être générée par HttpClient :
		- CurlException : Encapsule des erreurs Curl et des errno de celui-ci.
		- CurlLibException : Encapsule des erreurs d'utilisation la classe HTTPClient.

		La classe HttpClient utilise la librairie curl. Disponible via le packet php-curl sous Debian.
    @author Julien George
*/


/**
    @brief Classe pour faire du CURL en POO
    @details
		Cette classe encapsule les appelle à curl et transforme. On peut donc utiliser
		Celui-ci en programmation objet. et utiliser des méthodes get/post de manière
		simplifiée.
    classe qui étends celle-ci.
*/
class HttpClient
{
	/** @brief Un gestionnaire cURL retourné par curl_init(). */
	protected $curl=null;
	/** @brief Un tableau spécifiant quelles options à fixer avec leurs valeurs.*/
	protected $options=null;
	/** @brief L'url de du serveur web qui sera interrogé.*/
	protected $url=null;

	/** Contient NULL ou un ableau associatif avec des infos sur le dernier transfert cURL */
	protected $curl_info=NULL;
	/**
			@brief  Le seul constructeur, l'url du serveur web est obligatoire
			@details
			Par défaut, le constructeur configure :
			- les TIMEOUT sur 5 secondes
			- Désactive la vérification SSL
			- CURLOPT_FOLLOWLOCATION, CURLOPT_RETURNTRANSFER, CURLOPT_NOSIGNAL et CURLOPT_FRESH_CONNECT à TRUE.
			@param string $base_url L'url du serveur web qui sera interrogé
			@param boolean $cookies Permet d'activer de manière simple la gestion des cookies
	*/
	public function __construct($base_url,$cookies=FALSE){
		$this->url = $base_url;

		$options = array();
		$options[CURLOPT_FOLLOWLOCATION]=TRUE;
		$options[CURLOPT_RETURNTRANSFER]=TRUE;
		$options[CURLOPT_NOSIGNAL]=TRUE;
		$options[CURLOPT_FRESH_CONNECT]=TRUE;
		$options[CURLOPT_SSL_VERIFYPEER]=FALSE;
		$options[CURLOPT_SSL_VERIFYHOST]=FALSE;

		$options[CURLOPT_CONNECTTIMEOUT]=5;
		$options[CURLOPT_TIMEOUT]=5;

		$this->options=$options;
		$this->setCookies($cookies);
	}

	/**
			@brief  Initialise curl
			@details
			Initialise une nouvelle session CURL via curl_init et configure les options
	*/
	public function open(){
		if( $this->curl != null ) throw new CurlLibException('Curl resource already opened');
		$this->curl = curl_init();
		curl_setopt_array($this->curl ,$this->options);
	}

	/**
			@brief Active/configure les cookies
			@details
			Agit, configure les options suivante :
			- CURLOPT_COOKIEJAR
			- CURLOPT_COOKIEFILE
			- CURLOPT_COOKIESESSION

			Avec la valeur :
			- FALSE désactive les cookies
			- TRUE active les coolies en générant un fichier temporaire commençant par cookie_
			- un path, active le cookie et stocke dans le fichier temporaire $tmp_file
			@param bool|string $tmp_file TRUE|FALSE ou nom du fichier cookie.
	*/
  public function setCookies($tmp_file){
		if( !is_bool( $tmp_file ) || $tmp_file ){
			if( is_bool($tmp_file)){
				$tmp_file = tempnam( sys_get_temp_dir(), 'cookie_' );
			}
			$options[CURLOPT_COOKIEJAR] = $tmp_file;
			$options[CURLOPT_COOKIEFILE] = $tmp_file;
		}
		else{
			$options[CURLOPT_COOKIESESSION]=FALSE;
		}
		$this->setOptions($options);
        }

	/**
			@brief Configure l'objet avec un autre CurlHandle
			@param resource $curl Le CurlHandle
	*/
	public function setCurl(resource $curl){
		$this->curl = $curl;
	}

	/**
			@brief Retourne le CurlHandle
			@retval resource Le CurlHandle
	*/
	public function getCurl(){
		return $this->curl;
	}

	/**
			@brief Configure les options de Curl
			@details
			Peux s'utiliser de 2 manières différents :
			- setoptions(array $options); Pour fixer plusieurs options pour un transfert cURL
			- setoptions(string $option_key, mixed $value) pour fixer une option
			@param array|string $options_or_key Un tableau spécifiant quelles options à fixer avec leurs
				valeurs. Ou un string avec le nom d'une option.
			@param mixed $value La valeur à définir pour option.
	*/
	public function setoptions($options_or_key,$value=NULL){
		if( is_array($options_or_key)){
			foreach( $options_or_key as $k => $v ){
				$this->options[$k] = $v;
			}
		}
		else{
			$this->options[$options_or_key] = $value;
			$options_or_key = array($options_or_key=>$value);
		}

		if( $this->curl != null ) curl_setopt_array($this->curl,$options_or_key);
	}

	/**
		@brief Méthode générique pour exécuter une session de transfert cURL
		@details
		Cette fonction doit être appelée après l'appel d'une fonction suivante :
		- setCurl()
		- open()
		Afin d'avoir un ojet curl initialisé.
		@param array $options Un tableau spécifiant quelles options à fixer avec
			leurs valeurs avant l'exécutiob du transfert. Ces options ne sont pas
			enregistrées et ne conçerne dont qu'un seul transfert.
		@retval mixed par défaut, vour option CURLOPT_RETURNTRANSFER=TRUE,
			retourne le résultat en cas de succès, et false en cas d'échec.
	*/
	public function execute_request($options=null){
		if( $this->curl == null ) throw new CurlLibException('Curl resource no present. You must first calling open() method');
		$ch = curl_copy_handle($this->curl);

		if( $options != NULL ){
			curl_setopt_array($ch ,$options);
		}
		$content = curl_exec($ch);

		$errno = curl_errno($ch);
		if($errno){
			throw new CurlException(curl_strerror($errno),$errno);
		}

		$this->curl_info = curl_getinfo($ch);

		curl_close($ch);

		return $content;
	}

	/**
		@brief Lit les informations détaillant le dernier transfert cURL
		@retval array Un tableau associatif. Voir https://www.php.net/manual/fr/function.curl-getinfo.php
	*/
	public function getLastTransferInfo(){
                return $this->curl_info;
        }

	/**
		@brief retourne le code http de la réponse du dernier transfert cURL
		@retval integer Un code HTTP différent de 418 (I’m a teapot)
	*/
	public function getHttpCode(){
		return $this->curl_info['http_code'];
	}

	/**
		@brief ajoute un chemin à une url et/ou des paramètres GET
		@details
		- Récupére l'url configurée dans l'objet
		- Ajoute un chemin et/ou des paramètres GET
		- Retourne la nouvelle url
		@params string $path Le chemin souhaité
		@params array Un tableau de paramètres GET a passer à http_build_query (Voir https://www.php.net/manual/fr/function.http-build-query.php)
		@retval L'url de l'objet avec un chemin et/ou des paramètres GET
	*/
	private function build_url($path,$params=NULL){
		if( empty ($path) ) {
			$url = $this->url;
		}
		else{
			$url = rtrim($this->url,'/').'/'.ltrim($path,'/');
		}
		if( $params != NULL ){
			$data = http_build_query($params);
			$url.="?$data";
		}
		return $url;
	}

	/**
			@brief Exécute une requête avec la méthode GET
			@details
			Initialise Curl pour exécuter une méthode GET et l'exécute.
			Retourne le contenue de la réponse.
			@param string $path Le chemin souhaité à ajouter à l'url
			@param array $params Un tableau de paramètres GET sous forme array('key'=>'value');
			@param boolean $url_mode TRUE si on passe les paramètres dans l'URL.
				FALSE si on souhaite le faire comme avec la méthode POST
			@retval mixed par défaut, vour option CURLOPT_RETURNTRANSFER=TRUE,
				retourne le résultat en cas de succès, et false en cas d'échec.
	*/
	public function get($path,$params=null,$url_mode = TRUE){
		if( !$url_mode ){
			$options[CURLOPT_POSTFIELDS] = http_build_query($params);
			$params == NULL;
		}
		$options[CURLOPT_URL] = $this->build_url($path,$params);
		$response = $this->execute_request($options);
		return $response;
	}

	/**
			@brief Exécute une requête avec la méthode POST
			@details
			Initialise Curl pour exécuter une méthode POST et l'exécute.
			Retourne le contenu de la réponse.
			@param string $path Le chemin souhaité à ajouter à l'url
			@param array $params Un tableau de paramètres POST sous forme array('key'=>'value');
			@retval mixed par défaut, vour option CURLOPT_RETURNTRANSFER=TRUE,
				retourne le résultat en cas de succès, et false en cas d'échec.
	*/
	public function post($path,$params=null){
		$options[CURLOPT_URL] = $this->build_url($path);
		if( $params != NULL ){
			$data = http_build_query($params);
			$options[CURLOPT_POSTFIELDS]=$data;
		}
		$options[CURLOPT_POST]=TRUE;
		$response = $this->execute_request($options);
		return $response;
	}

	/**
			@brief Exécute une requête avec la méthode PUT
			@details
			Initialise Curl pour exécuter une méthode PUT et l'exécute.
			Retourne le contenu de la réponse.
			@param string $path Le chemin souhaité à ajouter à l'url
			@param array $params Un tableau de paramètres PUT sous forme array('key'=>'value');
			@retval mixed par défaut, vour option CURLOPT_RETURNTRANSFER=TRUE,
				retourne le résultat en cas de succès, et false en cas d'échec.
	*/
	public function put($path,$params=null){
		$options[CURLOPT_URL] = $this->build_url($path);
		$options[CURLOPT_CUSTOMREQUEST]='PUT';
                if( $params != NULL ){
                        $data = http_build_query($params);
                        $options[CURLOPT_POSTFIELDS]=$data;
                }

		$response = $this->execute_request($options);
		return $response;
	}

	/**
			@brief Exécute une requête avec la méthode DELETE
			@details
			Initialise Curl pour exécuter une méthode DELETE et l'exécute.
			Retourne le contenu de la réponse.
			@param string $path Le chemin souhaité à ajouter à l'url
			@retval mixed par défaut, vour option CURLOPT_RETURNTRANSFER=TRUE,
				retourne le résultat en cas de succès, et false en cas d'échec.
	*/
	public function delete($path){
		$options[CURLOPT_URL] = $this->build_url($path);
		$options[CURLOPT_CUSTOMREQUEST]='DELETE';
		$response = $this->execute_request($options);
		return $response;
	}

	/**
		@brief Ferme une session CURL
		@details
		Ferme une session cURL et libère toutes les ressources réservées.
 */
	public function close(){
		if( $this->curl != null ){
			curl_close($this->curl);
		}
		$this->curl = NULL;
	}
}

/** @brief Encapsule des erreurs Curl et des errno de celui-ci. */
class CurlException extends Exception {};

/** @brief Encapsule des erreurs d'utilisation la classe HTTPClient. */
class CurlLibException extends Exception {};
