<?php
/**
    @file
    @brief Contient les classes et méthodes que doivent implémenter tous driveur.
    @details
    Ce fichier une liste de classes abstraites ou pas qu'il est nécessaire
    d'étendre lors de la création d'un nouveau driveur.
    La classe principale, Cluster, permet de récupéer un objet principal qui
    est manipulé par les développeurs afin de récupérer d'autres objets
    instanciés par les différentes classes de ce fichier.
    @author Julien George
*/
if( !class_exists('Device') ){
    include_once 'Device.class.php';
}

/**
    @brief Classe principale du projet
    @details
    Cette classe permet aux utilisateurs d'instancier un objet de type
    Cluster en passant les paramètres d'authentication nécessaires pour
    accéder à l'API d'un système de virtualisation. Toutes les autres objets
    de l'API sont manipulées et distribués par des objets instanciant une
    classe qui étends celle-ci.
*/
abstract class Cluster implements JsonSerializable{
    /**
        @brief Constructeur avec les paramètres d'authentification
        @details
        Toutes classes étendant celle-ci doivent fournir un Constructeur
        permettant l'initialisation des paramètres d'authentification.
        @param array $settings Paramètres d'authenficiation nécessaire pour la
            connection à l'API. Le tableau contient toujours les valeurs :
                Array('url'=>"scheme://hostname:port",'login'=>"",password=>"")
*/
    public function __construct($settings){}

    /**
        @brief Retourne le nom du cluster.
        @retval string Le nom du Cluster
        @retval NULL Si le nom n'est pas dispo. (Hyperviseur standalone par exemple).
    */
    abstract function getName();

    /**
        @brief Retourne le nom de l'hyperviseur maître
        @retval string Le nom du Cluster
        @retval NULL Si le nom n'est pas dispo.
    */
    abstract function getMasterHost();

    /**
        @brief Indique si nous avont à faire à un cluster ou un hyperviseur seul.
        @retval boolean
    */
    abstract function isCluster();

    /**
        @brief Indique si la HA est active et fonctionne
        @retval boolean
    */
    abstract function isHA();

    /**
        @brief Retourne la liste de tous les hyperviseurs
        @retval array<Host> Une liste d'objets Host
    */
    abstract function getHosts():array;

    /**
        @brief Retourne la liste de toutes les VMs
        @retval array<VM> Une liste d'objets VM (excluant les templates)
    */
    abstract function getVMs():array;

    /**
        @brief Retourne la liste de toutes les Templates
        @retval array<VM> Une liste de Templates
    */
    abstract function getTemplates():array;

    /**
        @brief Retourne la liste de toutes les Stockages
        @param boolean|NULL $shared
            - Si NULL, tous les stockages
            - Si TRUE, uniquement les stockages partagés
            - Si FALSE, uniquement les stockages locaux.
        @retval array<Storage> Une liste d'objets Storage
    */
    abstract function getStorages($shared=NULL):array;

    /**
        @brief Retourne les stockages dont le nom est précisé
        @remark Cette fonction est déjà définie
        @param string $name Le nom du (ou des si il en a plusieurs) Storage
        @retval array<Storage> Une liste d'objets Storage (souvent 1 seul)
    */
    function getStorageByName($name){
        $ret = array();
        $all = $this->getStorages();
        foreach($all as $sr){
            if( $sr->getName() == $name){
                $ret[] = $sr;
            }
        }
        return $ret;
    }

    /**
        @brief Retourne l'hyperviseur dont le nom est précisé
        @param string $hostname Le nom de l'hyperviseur
        @retval Host Un objet de classe Host
    */
    abstract function getHostByName($hostname);

    /**
        @brief Permet de trouver une VM dont l'ID du snapshot est indiqué
        @param $snapid ID permettant de retrouver un snapshot quelconque
        @retval VM Un objet de classe VM contenant la VM conçernée par le snapshot.
    */
    abstract function getVMFromSnapshot($snapid);

    /**
        @brief Retourne La VM dont le nom est précisé
        @param string $name Le nom de la VM
        @retval Host Un objet de classe VM
    */
    abstract function getVMByName($name);

    /**
        @brief Retourne La VM dont d'ID est spécifié
        @param string $vmid l'ID de la VM
        @param boolean $refresh Avec $refresh par défault = FALSE, on n'interroge pas l'API,
            on passe par un cache. TRUE efface le cache et oblige donc à passer par l'API.
        @retval VM Un objet de classe VM
    */
    abstract function getVM($vmid,$refresh=FALSE);

    /**
        @brief Permet de cloner une VM ou un template
        @param string|array|VM $source La source, sous la forme d'un objet VM,
            un array['id'] ou un string composé de l'id de la VM
        @param string $target_name Nom qui sera donné à la nouvelle VM
        @param boolean $full TRUE par défaut, indique de faire un clonage complet
    */
    function cloneVM($source,$target_name,$full=TRUE){
        throw new Exception('Not implemented');
    }

    /**
        @brief Retourne l'utilisation totale CPU des VMs possédant la HA
        @remark Cette fonction est déjà définie
        @retval float une valeur comprise en 0 et 1 pouvant être transformé en %
    */
    public function getCpusHA(){
        $cl_used_cpus = 0;
        $cl_total = 0;
        foreach($this->getHosts() as $host){
            if( $host->isAlive()){
                $host_total = $host->getCpusTotal();
                $host_used_cpus = $host_total * $host->getCpusHA();

                $cl_used_cpus += $host_used_cpus;
                $cl_total += $host_total;
            }
        }
        return $cl_used_cpus/$cl_total;
    }

    /**
        @brief Retourne l'utilisation totale CPU des VMs avec l'état running
        @remark Cette fonction est déjà définie
        @retval float une valeur comprise en 0 et 1 pouvant être transformé en %
    */
    public function getCpusUsage(){
        $cl_used_cpus = 0;
        $cl_total = 0;
        foreach($this->getHosts() as $host){
            if( $host->isAlive()){
                $host_total = $host->getCpusTotal();
                $host_used_cpus = $host_total * $host->getCpusUsage();

                $cl_used_cpus += $host_used_cpus;
                $cl_total += $host_total;
            }
        }
        return $cl_used_cpus/$cl_total;
    }

    /**
        @brief Retourne le nombre total de CPU disponibles dans le cluster
        @remark Cette fonction est déjà définie
        @retval integer Le nombre total de CPU dans le cluster
    */
    public function getCpusTotal(){
        $ret = 0;
        foreach($this->getHosts() as $host){
            if( $host->isAlive()){
                $ret += $host->getCpusTotal();
            }
        }
        return $ret;
    }

    /**
        @brief Retourne l'utilisation totale mémoire des VMs possédant la HA
        @remark Cette fonction est déjà définie
        @retval integer Le nombre d'octets utilisés
    */
    public function getMemHa(){
        $ret = 0;
        foreach($this->getHosts() as $host){
            if( $host->isAlive()){
                $ret += $host->getMemHa();
            }
        }
        return $ret;
    }

    /**
        @brief Retourne l'utilisation totale mémoire des VMs avec l'état running
        @remark Cette fonction est déjà définie
        @retval integer Le nombre d'octets utilisés
    */
    public function getMemUsage(){
        $ret = 0;
        foreach($this->getHosts() as $host){
            if( $host->isAlive()){
                $ret += $host->getMemUsage();
            }
        }
        return $ret;
    }

    /**
        @brief Retourne la capacité totale actuelle du cluster
        @remark Cette fonction est déjà définie
        @retval integer Le nombre d'octets
    */
    public function getMemTotal(){
        $ret = 0;
        foreach($this->getHosts() as $host){
            if( $host->isAlive()){
                $ret += $host->getMemTotal();
            }
        }
        return $ret;
    }

    /**
        @brief Retourne le traffic total entrant dans les hyperviseurs du cluster
        @remark Cette fonction est déjà définie
        @retval integer Le nombre de bits/s
    */
    function getNetInUsage(){
        $ret = 0;
        foreach($this->getHosts() as $host){
            if( $host->isAlive()){
                $ret += $host->getNetInUsage();
            }
        }
        return $ret;
    }

    /**
        @brief Retourne le traffic total sortant des hyperviseurs du cluster
        @remark Cette fonction est déjà définie
        @retval integer Le nombre de bits
    */
    function getNetOutUsage(){
        $ret = 0;
        foreach($this->getHosts() as $host){
            if( $host->isAlive()){
                $ret += $host->getNetOutUsage();
            }
        }
        return $ret;
    }

    /*
        format:NULL = default
        format:'ID' = getID()
        format:'NAME = getNAME()
    */

    /**
        @brief Permet d'accèder aux donées de l'objet en utilisant des $keys.
        @details
        Permet d'accèder aux données de l'objet sans passer par des méthodes.
        Peut donc être utiliser dans des boucles ou autre choses.
        @remark Cette fonction est déjà définie
        @param string $key C'est le nom de la donnée à récupérer.
        @param NULL|string $format Si $format vaut id ou name, transformera la
            donnée renvoyée en id ou name ou liste d'id/name.
            Par défaut = NULL ne transformera pas la donnée.
    */
    public function getField($key,$format=NULL){
        $key = strtolower($key);
        $ret = NULL;
        if    ( $key == 'name')         $ret = $this->getName();
        elseif( $key == 'master_host')  $ret = $this->getMasterHost();
        elseif( $key == 'is_cluster')   $ret = $this->isCluster();
        elseif( $key == 'is_ha')        $ret = $this->isHA();

        elseif( $key == 'hosts')        $ret = $this->getHosts();
    //  elseif( $key == 'storages')     $ret = $this->getStorages();
        elseif( $key == 'templates')    $ret = $this->getTemplates();
        elseif( $key == 'vms')          $ret = $this->getVMs();

        elseif( $key == 'cpus_ha')      $ret = $this->getCpusHA();
        elseif( $key == 'cpus_usage')   $ret = $this->getCpusUsage();
        elseif( $key == 'cpus_total')   $ret = $this->getCpusTotal();
        elseif( $key == 'mem_ha')       $ret = $this->getMemHa();
        elseif( $key == 'mem_usage')    $ret = $this->getMemUsage();
        elseif( $key == 'mem_total')    $ret = $this->getMemTotal();
        elseif( $key == 'net_in')       $ret = $this->getNetInUsage();
        elseif( $key == 'net_out')      $ret = $this->getNetOutUsage();

        return cluster::formatField($ret,$format);
    }

    /**
        @brief Transforme une donnée objet ou array en id ou name
        @details
        Cette fonction a pour but de transformer une donnée brut en ID ou en
        name.
        - Si $data contient un objet ou un array, elle retournera l'id ou le
        name de cette objet.
        - Si $data contient un tableau d'objet ou d'array, ce sera un Tableau
        d'id ou de name qui sera retourné.
        - Si $data contient autre chose, celui-ci sera retourné tel quel.
        Si $key vaut null, alors $data sera toujours retourné tel quel.
        @remark Cette fonction est déjà définie
        @remark Une autre key que id ou name peut être spécifiée.
        @param mixed $data Une ou plusiers données qui seront transformés ou pas
            en id ou name
        @param string $key Si NULL, on retourne la donnée tel quel, sinon, on
            retourne la valeur de $key
        @retval string|array<string>|mixed Un id ou name, Un tableau (id ou name) ou la donnée en brut.
    */
    public static function formatField($data,$key){
        if( $key != NULL ){
            $isArray = TRUE;
            if( is_object($data)){
                    $data = array($data);
                    $isArray = FALSE;
            }
            elseif( is_array($data) && !isset($data[0])) {    //Tableau associatif
                $data = array($data);
                $isArray = FALSE;
            }
            if( is_array($data)){
                $tmp = array();
                foreach($data as $row){
                    if(is_object($row)){
                        $tmp[] = $row->getField($key);
                    }
                    elseif(is_array($row) &&!empty($row)){
                        if( !isset($row[$key])) print_r($row);
                        $tmp[] = $row[$key];
                    }
                    else break;
                }
                if(!$isArray) $tmp = array_shift($tmp);
                $data = $tmp;
            }
        }
        return $data;
    }

    /**
        @brief Permet de transformer un objet en un array
        @remark Cette fonction est déjà définie
        @param object $obj L'objet à transformer en array
        @param array $basics La liste des clés retournant une valeur basique
        @param array $lists La liste des clés retournant un tableau
        @param boolean $human Convertit les valeurs en valeur lisible par les humains.
          Si ce paramètre est à TRUE, alors les valeurs retournées seront :
            - .Le name au lieu de l'id
            - 'N/A' sera retournée à la place d'une valeur NULL
            - Les valeurs pouvant être converties en pourcents seront converties en XX%
            - Les mémoires ram et disques seront converties en MB ou GB
            - Les vitesses seront converties en Mb/s ou MB/s
            - Les uptimes seront converties en X days x hours ...
        @retval array Un tableau associatif contenant les clés de $basics et $lists
    */
    public static function convertObjectToArray($obj,$basics,$lists,$human){
        $ret = array();
        if( $human ){
            $format = 'name';
        }
        else{
            $format = 'id';
        }

        foreach($basics as $key){
            $ret[$key] = $obj->getField($key,$format);
        }

        if( $human){
            foreach($ret as $key => $row){
                if( is_null($row)){
                    $ret[$key] = 'N/A';
                }
                elseif( is_bool($row)){
                    if($row)    $ret[$key] = 'TRUE';
                    else        $ret[$key] = 'FALSE';
                }
                elseif( is_numeric($row)){
                    $type = explode('_',$key);
                    if( $type[0] == 'cpus' && $type[1] != 'total'){
                        $row = $row*100;
                        $row = round($row,1);
                        $ret[$key] = "$row%";
                    }
                    elseif( $type[0] == 'mem'){
                        if( is_subclass_of($obj,'host') ){
                            $row = $row /1024/1024/1024;
                            $row = round($row,1);
                            $ret[$key] = "$row GB";
                        }
                        else{
                            $row = $row /1024/1024;
                            $row = round($row,1);
                            $ret[$key] = "$row MB";
                        }
                    }
                    elseif( $type[0] == 'size' ){
                        $row = $row /1024/1024/1024;
                        $row = round($row,1);
                        $ret[$key] = "$row GB";
                    }
                    elseif($type[0] == 'net' ){
                        $row = $row /1000/1000;
                        $row = round($row,1);
                        $ret[$key] = "$row Mb/s";
                    }
                    elseif($type[0] == 'disk'){
                        $row = $row /1024;
                        $row = round($row,1);
                        $ret[$key] = "$row KB/s";
                    }
                    elseif($type[0] == 'uptime'){
                        $ut = $row;

                        $days = round($ut/(3600*24),0,PHP_ROUND_HALF_DOWN);
                        $ut %=(3600*24);

                        $hours = round($ut/3600,0,PHP_ROUND_HALF_DOWN);
                        $ut %=3600;

                        $minutes =  round($ut/60,0,PHP_ROUND_HALF_DOWN);
                        $ut %=60;
                        $secondes = $ut;

                        $ret[$key] =  $days." days ".$hours." hours ".$minutes." minutes ".$secondes." secondes";
                    }
                    elseif( is_object($row)){
                        $ret[$key] = $row->getName();
                    }
                }
            }
        }

        foreach( $lists as $key){
            $ret[$key] = $obj->getField($key,$format);
        }

        return $ret;
    }

    /**
        @brief Transforme un objet en un array
        @remark Cette fonction est déjà définie
        @details
        Cette fonction peut être utilisée sans paramètres. Ou on peut préciser
        Si on souhaite récupérer les métriques et les listes.
        Généralement, la récupération de metriques prends du temps à travers
        les APIs hyperviseurs, pareils pour les listes.
        Si vous souhaitez choisir les champs à récupéer, utiliser plutot
        convertObjectToArray()
        - Les champs basiques retournés sont : "name,master_host,is_cluster,is_ha"
        - Les champs metriques retournés sont : "cpus_ha,cpus_usage,cpus_total,mem_ha,mem_usage,mem_total,net_in,net_out"
        - Les champs listes retournés sont : "hosts,storages,templates,vms"
        @param boolean $enable_metrics Retourne aussi les données de type metrics (plus lent)
        @param boolean $enable_lists Retourne aussi les données de type list (plus lent)
        @param boolean $human FALSE pour des données bruts. TRUE pour des données 'compréhensibles'.
        @retval array Un tableau associatif contenant les clés de $basics et $lists
    */
    public function toArray($enable_metrics=FALSE,$enable_lists=FALSE,$human=FALSE){
        $basics     = explode(',','name,master_host,is_cluster,is_ha');

        $lists = $from_list = array();
        if( $enable_lists   )   $lists     = explode(',','hosts,storages,templates,vms');
        if( $enable_metrics )   $from_list  = explode(',','cpus_ha,cpus_usage,cpus_total,mem_ha,mem_usage,mem_total,net_in,net_out');

        $basics = array_merge($basics,$lists,$from_list);
        return cluster::convertObjectToArray($this,$basics,$lists,$human);
    }

    /**
        @brief Implémentation de l'interface JsonSerializable
        @details
        Permet de serialiser en json l'objet. De manière à l'envoyer directement
        à la méthode json_encode. Utilise la fonction toArray() sans paramètres et donc
        ne sérialise que les informatiques basiques (pas de métriques ou list).
    */
    public function jsonSerialize(){
        return $this->toArray();
    }
}

/**
    @brief Classe abstraite contenant les méthodes communes à Host et VM
    @details
    Cette classe étends la classe Device commune a tous les noeuds Esia.
    Dans le cadre d'une utilisation sans l'interface d'Esia, un fichier Device.class.php
    situé dans le même répertoire que ce fichier est inclut. Ce fichier Contient
    une classe Device identique ou presque à celle de l'interface d'Esia.
    @remark Certaines méthodes doivent être rédéfinies.
*/
abstract class CommonVirtDevice extends Device implements JsonSerializable{
   /**
      @brief Constructeur par défaut identique à celui de Device d'Esia.
      @param numeric|null $id Un identifiant unique
   */
    public function __construct ($id=NULL){
        $this->setId($id);
    }

    /**
       @brief Fonction de configuration de l'adresse IPv4
       @details
       Détecte si c'est un long IPv4 ou un string IPV4 et appelle la bonne méthode
       de configuration.
       @param integer|string  $ip Soit un longIPv4 soit un StringIPv4
    */
    public function setAddress($ip){
        if( is_numeric($ip) ){
            $this->setIPv4DecimalAdress($ip);
        }
        else{
            $this->setIPv4Adress($ip);
        }
    }

    /**
       @brief Récupére l'adresse IP du device
       @retval string Un StringIPv4
    */
    public function getAddress(){
        return (string)$this->getIPv4Adress();
    }

    /**
       @brief Récupére l'uptime de la machine (Nombre de secondes depuis le boot)
       @retval integer L'uptime en nombre de secondes
    */
    public function getUpTime(){
        return NULL;
    }

    /**
       @brief Retourne le nombre total de core CPUs disponibles.
       @retval integer Le nombre total de CPUs
    */
    function getCpusTotal(){
        return NULL;
    }

    /**
       @brief Retourne l'utilisation totale CPUs
       @retval float	une valeur comprise en 0 et 1 pouvant être transformé en %
    */
    function getCpusUsage(){
        return NULL;
    }

    /**
       @brief Retourne l'utilisation totale mémoire.
       @retval integer	Le nombre d'octets disponibles
    */
    function getMemTotal(){
        return NULL;
    }

    /**
       @brief Retourne l'utilisation mémoire.
       @retval integer Le nombre d'octets utilisés
    */
    function getMemUsage(){
        return NULL;
    }

    /**
       @brief Retourne le traffic total entrant.
       @retval integer	Le nombre de bits/s
    */
    function getNetInUsage(){
        return NULL;
    }

    /**
       @brief Retourne le traffic total sortant.
       @retval integer	Le nombre de bits/s
    */
    function getNetOutUsage(){
        return NULL;
    }

    /**
       @brief Indique si la machine est opérationnel ou pas (Démarrée/accessible)
       @retval boolean Retourne TRUE si la machine est opérationnelle
    */
    function isAlive(){
        return NULL;
    }

    /**
        @brief Permet d'accèder aux donées de l'objet en utilisant des $keys.
        @details
        Permet d'accèder aux données de l'objet sans passer par des méthodes.
        Peut donc être utiliser dans des boucles ou autre choses.
        @remark Cette fonction est déjà définie
        @param string $key C'est le nom de la donnée à récupérer.
        @param NULL|string $format Si $format vaut id ou name, transformera la
            donnée renvoyée en id ou name ou liste d'id/name.
            Par défaut = NULL ne transformera pas la donnée.
    */
    public function getField($key,$format=NULL){
        $key = strtolower($key);
        $ret = NULL;
        if    ( $key == 'id')            $ret = $this->getID();
        elseif( $key == 'name')         $ret = $this->getName();
        elseif( $key == 'description')  $ret = $this->getDescription();
        elseif( $key == 'type')         $ret = $this->getType();
        elseif( $key == 'manufacturer') $ret = $this->getManufacturer();
        elseif( $key == 'os')           $ret = $this->getOS();
        elseif( $key == 'os_version')   $ret = $this->getOSVersion();
        elseif( $key == 'address')      $ret = $this->getAddress();
        elseif( $key == 'uptime')       $ret = $this->getUpTime();
        elseif( $key == 'alive')        $ret = $this->isAlive();

        elseif( $key == 'cpus_usage')   $ret = $this->getCpusUsage();
        elseif( $key == 'cpus_total')   $ret = $this->getCpusTotal();
        elseif( $key == 'mem_usage')    $ret = $this->getMemUsage();
        elseif( $key == 'mem_total')    $ret = $this->getMemTotal();
        elseif( $key == 'net_in')       $ret = $this->getNetInUsage();
        elseif( $key == 'net_out')      $ret = $this->getNetOutUsage();

        return cluster::formatField($ret,$format);
    }

    /**
        @brief Transforme un objet en un array
        @remark Cette fonction est déjà définie
        @details
        Cette fonction peut être utilisée sans paramètres. Ou on peut préciser
        Si on souhaite récupérer les métriques et les listes.
        Généralement, la récupération de metriques prends du temps à travers
        les APIs hyperviseurs, pareils pour les listes.
        Si vous souhaitez choisir les champs à récupéer, utiliser plutot
        convertObjectToArray()
        - Les champs basiques retournés sont : "id,name,description,type,manufacturer,os,os_version,address,uptime,alive"
        - Les champs metriques retournés sont : "cpus_usage,cpus_total,mem_usage,mem_total,net_in,net_out"
        - Les champs listes retournés sont : Il n'y en pas.
        @param boolean $enable_metrics Retourne aussi les données de type metrics (plus lent)
        @param boolean $enable_lists Retourne aussi les données de type list (plus lent)
        @param boolean $human FALSE pour des données bruts. TRUE pour des données 'compréhensibles'.
    */
    public function toArray($enable_metrics=FALSE,$enable_lists=FALSE,$human=FALSE){
        $basics     = explode(',','id,name,description,type,manufacturer,os,os_version,address,uptime,alive');

        $lists = $from_list = array();
        if( $enable_metrics )   $from_list  = explode(',','cpus_usage,cpus_total,mem_usage,mem_total,net_in,net_out');

        $basics = array_merge($basics,$lists,$from_list);
        return cluster::convertObjectToArray($this,$basics,$lists,$human);
    }

    /**
        @brief Implémentation de l'interface JsonSerializable
        @details
        Permet de serialiser en json l'objet. De manière à l'envoyer directement
        à la méthode json_encode. Utilise la fonction toArray() sans paramètres et donc
        ne sérialise que les informatiques basiques (pas de métriques ou list).
    */
    public function jsonSerialize(){
        return $this->toArray();
    }
}

/**
  @brief Représente un hyperviseur. Doit être étendue par les classes des Driveurs.
  @remark Certaines méthodes de Host et CommonVirtDevice doivent être redéfinies.
*/
class Host extends CommonVirtDevice{
    /**
      @brief retourne la liste des VMs de l'hypervisuer
      @retval array<VM>	Une liste d'objets VM (excluant les templates)
    */
    public function getVMs(){
            return NULL;
    }

    /**
      @brief retourne la liste des VMs de l'hypervisuer
      @retval array<VM>	Une liste de Templates
    */
    function getTemplates(){
        return NULL;
    }

    /**
      @brief Retourne des informations globales réseau et des info spécifique aux interfaces
      @retval Le nombre de bits/s
    */
    public function getNetInterfaces()
    {
            return NULL;
    }

    /**
        @brief Retourne l'utilisation totale CPU des VMs possédant la HA
        @remark Cette fonction est déjà définie
        @retval float une valeur comprise en 0 et 1 pouvant être transformé en %
    */
    public function getCpusHA(){
        $result_core_total = $this->getCpusTotal();
        $all_core_total = 0;
        $all_core_usage = 0;

        $list = array();
        foreach($this->getVMs() as $row){
            if( $row->isAlive() && $row->isHA()){
                $list[] = $row;
            }
        }
    //    $list[] = $this; //Le Host utilise aussi du CPU

        if( empty($list)) return 0;

        foreach($list as $row){
                $row_core_total = $row->getCpusTotal();
                $row_core_percent = $row->getCpusUsage();
                $row_core_usage = $row_core_total * $row_core_percent;

                $all_core_total += $row_core_total;
                $all_core_usage += $row_core_usage;
        }
        $all_core_percent = $all_core_usage / $all_core_total;
        $result_percent_usage = ($all_core_percent / $result_core_total) * $all_core_total;

        return $result_percent_usage;
    }

    /**
        @brief Retourne l'utilisation totale mémoire des VMs possédant la HA
        @remark Cette fonction est déjà définie
        @retval integer Le nombre d'octets utilisés
    */
    public function getMemHA(){
        $ret = 0;
        foreach($this->getVMs() as $vm){
            if( $vm->isHA()){
                $ret += $vm->getMemUsage();
            }
        }
        return $ret;
    }

    /**
        @brief Permet d'accèder aux donées de l'objet en utilisant des $keys.
        @details
        Permet d'accèder aux données de l'objet sans passer par des méthodes.
        Peut donc être utiliser dans des boucles ou autre choses.
        @remark Cette fonction est déjà définie
        @param string $key C'est le nom de la donnée à récupérer.
        @param NULL|string $format Si $format vaut id ou name, transformera la
            donnée renvoyée en id ou name ou liste d'id/name.
            Par défaut = NULL ne transformera pas la donnée.
    */
    public function getField($key,$format=NULL){
        $key = strtolower($key);
        $ret = NULL;
        if    ( $key == 'vms')            $ret = $this->getVMs();
        elseif( $key == 'templates')      $ret = $this->getTemplates();

        elseif( $key == 'cpus_ha')   $ret = $this->getCpusHA();
        elseif( $key == 'mem_ha')    $ret = $this->getMemHA();

        else return parent::getField($key,$format);

        return cluster::formatField($ret,$format);
    }

    /**
         @brief Transforme un objet en un array
         @remark Cette fonction est déjà définie
         @details
         Cette fonction peut être utilisée sans paramètres. Ou on peut préciser
         Si on souhaite récupérer les métriques et les listes.
         Généralement, la récupération de metriques prends du temps à travers
         les APIs hyperviseurs, pareils pour les listes.
         Si vous souhaitez choisir les champs à récupéer, utiliser plutot
         convertObjectToArray().
         - Les champs basiques retournés sont : "id,name,description,type,manufacturer,os,os_version,address,uptime,alive"
         - Les champs metriques retournés sont : "cpus_ha,cpus_usage,cpus_total,mem_ha,mem_usage,mem_total,net_in,net_out"
         - Les champs listes retournés sont : "vms,templates"
         @param boolean $enable_metrics Retourne aussi les données de type metrics (plus lent)
         @param boolean $enable_lists Retourne aussi les données de type list (plus lent)
         @param boolean $human FALSE pour des données bruts. TRUE pour des données 'compréhensibles'.
         @retval array Un tableau associatif contenant les clés de $basics et $lists
     */
    public function toArray($enable_metrics=FALSE,$enable_lists=FALSE,$human=FALSE){
        $basics     = explode(',','id,name,description,type,manufacturer,os,os_version,address,uptime,alive');

        $lists = $from_list = array();
        if( $enable_lists )     $lists  = explode(',','vms,templates');
        if( $enable_metrics )   $from_list  = explode(',','cpus_ha,cpus_usage,cpus_total,mem_ha,mem_usage,mem_total,net_in,net_out');

        $basics = array_merge($basics,$lists,$from_list);
        return cluster::convertObjectToArray($this,$basics,$lists,$human);
    }

    /**
        @brief Implémentation de l'interface JsonSerializable
        @details
        Permet de serialiser en json l'objet. De manière à l'envoyer directement
        à la méthode json_encode. Utilise la fonction toArray() sans paramètres et donc
        ne sérialise que les informatiques basiques (pas de métriques ou list).
    */
    public function jsonSerialize() {
            return $this->toArray();
    }
}

/**
  @brief Représente une Virtual Machine. Doit être étendue par les classes des Driveurs.
  @remark Certaines méthodes de VM et CommonVirtDevice doivent être redéfinies.
*/
class VM extends CommonVirtDevice{
   /**
     @brief Retourne des informations sur le CPU (architecture, modele, ...)
     @remark En fonction du système de virtualization, les valeurs retournées
      peuvent avoir une forme différente.
     @retval string Des infos CPU.
   */
    public function getCpuPlatform(){
        return NULL;
    }

    /**
      @brief Indique si l'agent de virtualisation est installé.
      @retval boolean TRUE si l'agent est activé.
    */
    public function isAgent(){
        return NULL;
    }

    /**
      @brief Indique le status démarré ou éteint de la VM
      @details
      Les status retournés sont toujours un de ceux-ci :
      - running : la vm est démarrée
      - paused : la vm est en pause
      - stopped : la vm est arrétée
      @retval string Une des valeurs suivantes : running, paused, stopped
    */
    public function getState(){
        return NULL;
    }

    /**
      @brief Indique si la VM est verrouillée
      @details
      Si cette méthode retourne TRUE, cela signifie que la VM est actuellement bloquée
      par une tâche du cluster, de l'hyperviseur (migration, clonage, ...).
      @retval boolean TRUE si la VM est verouillée. FALSE si elle est disponible.
    */
    public function isLocked($refresh=FALSE){
        return NULL;
    }

    /**
      @brief Retourne la raison du verrouillage (@see isLocked()) de la VM.
      @remark Le status/La raison du verrouillage dépend du système de virtualisation.
      @retval string la raison du verrou.
    */
    public function getLockStatus($refresh=FALSE){
        return NULL;
    }

    /**
      @brief Retourne l'hyperviseur de la VM
      @details
      Retourne l'hyperviseur lié à la VM.
      - Soit celui ou tourne cette VM.
      - Soit, pour une VM locale, celui on peut être démarré une VM.
      - Avec certains systèmes de virtualisation, une VM éteinte est lié à un
        hyperviseur à titre informatif (Elle peut être démarré sur les autres).
      @retval NULL|Host NULL si la VM n'est pas liée. Host si elle est liée à un hyperviseur.
    */
    function getHost(){
        return NULL;
    }

    /**
        @brief Indique si la HA est activée sur la VM. La VM est donc agile.
        @retval boolean Si la HA est activée.
    */
    function isHA(){
        return NULL;
    }

    /**
        @brief Retourne les informations conçernant la HA.
        @remark En fonction du système de virtualization, les valeurs retournées
         peuvent avoir une forme différente.
        @retval string Description de la configuration de la HA.
    */
    function getHASettings(){
        return NULL;
    }

    /**
        @brief Retourne les vitesses de lecture disque (ensemble des disques de la VM) en Bytes/s
        @retval integer Bytes/s
    */
    function getDiskRead(){
        return NULL;
    }

    /**
        @brief Retourne les vitesses d'écriture disque (ensemble des disques de la VM) en Bytes/s
        @retval integer Bytes/s
    */
    function getDiskWrite(){
        return NULL;
    }

    /**
        @brief Retourne une liste de tableau associatifs de snapshots.
        @details
        Ci-dessous un exemple de tableau snapshot
        ```php
          $snapshots_list = array();
          $snapshot['id']= $id;
          $snapshot['name']='Mon snapshot';
          $snapshot['description']='Ceci est mon snapshot';
          $snapshot['parent']= $parent;
          $snapshot['time']= $timestamp;
          $snapshot['withRAM']= FALSE;
          $snapshots_list[] = $snapshot;
          return $snapshots_list;
        ```
        Les éléments sont :
        - 'id' : L'id du snapshot
        - 'name' : Le nom du snapshot
        - 'description' : La description du snapshot
        - 'parent' : Le parent (VM ou autre snapshot). Cela doit être un ID. NULL si il n'y a pas de parent.
        - 'time' : Le timestamp conçernant la date/heure de création
        - 'withRAM' : à TRUE si le contenu de la RAM a été enregistrée dans le snapshot.
        @retval array<array> La liste de tableaux associatifs décrit ci-dessus.
    */
    function getSnapshots(){
        return NULL;
    }

    /**
        @brief Retourne une liste de tableau associatifs de volumes.
        @details
        Ci-dessous un exemple de tableau de volumes
        ```php
          $volumes_list = array();
          $volume['id']= $id:
          $volume['name']='Mon volume';
          $volume['desc']='Ceci est mon volume';
          $volume['size_used']= 8589934592; //8 GB
          $volume['size_virt']= 8589934592;
          $volume['parent']= NULL;
          $volume['storage']= $array_storages;
          $volume['vm'] = $array_vm;
          $volume['io_read']= 9127
          $volume['io_write']= 350
          $volumes_list[] = $volume;
          return $volumes_list;
        ```
        Les éléments sont :
        - 'id' : L'id du volume
        - 'name' : Le nom du volume
        - 'desc' : La description du volume
        - 'size_used' : La taille du disque sur le storage.
            Soit inférieure, soit identique à size_virt (allocation dynamique ou pas).
        - 'size_virt' : La taille allouée au storage.
        - 'parent' : L'id du volume parent si il y en a un. NULL sinon.
        - 'storage' : Une tableau contenant des objets Storage.
        - 'vm' : Un tableau contenant des objets VMs (C'est rare mais il peut y en avoir plusieurs)
        - 'io_read' : Vitesses de lecture sur ce volume en Bytes/s
        - 'io_write' : Vitesses d'écritue sur ce volume en Bytes/s
        @retval array<array> La liste de tableaux associatifs décrit ci-dessus.
    */
    function getVolumes(){
        return NULL;
    }

    /**
        @brief Crée un snapshot de la VM actuelle. Ou génére une exception en cas d'echec.
        @param string $name : Le nom du snapshot.
        @param string $description : Une description.
        @param boolean $withRam : TRUE par défaut. Indique si il faut enregistrer la RAM.
    */
    public function createSnapshot($name,$description=NULL,$withRam=NULL){
        throw new Exception('Not implemented');
    }

    /**
        @brief Supprime un snapshot de la VM actuelle. Ou génére une exception en cas d'echec.
        @param string $snapid : L'id du snapshot.
    */
    public function removeSnapshot($snapid){
        throw new Exception('Not implemented');
    }

    /**
        @brief Restaure le snapshot de la VM actuelle. Ou génére une exception en cas d'echec.
        @param string $snapid : L'id du snapshot.
    */
    public function reverseToSnapshot($snapid){
        throw new Exception('Not implemented');
    }

    /**
        @brief Démarre cette VM. Ou génére une exception en cas d'echec.
    */
    public function start(){
        throw new Exception('Not implemented');
    }

    /**
        @brief Redémarre cette VM. Ou génére une exception en cas d'echec.
        @param boolean $cleanly TRUE par défaut pour un redémarrage propre.
          Mettre à FALSE pour un hard reboot.
    */
    public function reboot($cleanly){
        throw new Exception('Not implemented');
    }

    /**
        @brief Arrête cette VM. Ou génére une exception en cas d'echec.
        @param boolean $cleanly TRUE par défaut pour un arrêt propre.
          Mettre à FALSE pour un hard shutdown.
    */
    public function shutdown($cleanly){
        throw new Exception('Not implemented');
    }

    /**
        @brief Migre cette VM. Ou génére une exception en cas d'echec.
        @param string $target Nom de l'hyperviseur
        @param boolean $online Si TRUE, tente de migrer la VM online.
          Mettre à FALSE pour un hard shutdown.
    */
    public function migrate($target,$online){
        throw new Exception('Not implemented');
    }

    /**
        @brief Clone cette VM. Ou génére une exception en cas d'echec.
        @param string $target_name Nom de la copie
        @param boolean $full TRUE par défaut, Fait une copie compléte.
    */
    public function cloneVM($target_name,$full=TRUE){
        throw new Exception('Not implemented');
    }

    /**
        @brief Attends qu'une VM se verrouille ou se déverouille.
        @details
        Attends et vérifie tous les $tick (en ms) que la VM ne soit plus verouillée (si $lock=FALSE).
        La durée d'attente est au maximum : $waitMin + $tick + Temps de réponse requête.
        @remark Cette fonction est déjà définie
        @param numeric $waitMin Nombre de secondes à attendre.
        @param boolean $lock Par défaut TRUE, pour attendre que la VM soit déverouillée.
        @param numeric $tick Temps en ms entre 2 vérification de status.
    */
    public function wait($waitMin,$lock=FALSE,$tick=600){
        $start = time();
        $end = $start + $waitMin;

        do{
            time_nanosleep(0, $tick*1000000);

            if( $lock == $this->isLocked(TRUE)){
                return TRUE;
            }

            $spended = time();
        }while($spended<$end);
        return FALSE;
    }

    /**
        @brief Retourne les différents types d'opérations possible pour une VM
          gérée par cette API.
        @details
        Les opérations possiblent connues par l'API sont :
        - start
        - clone
        - shutdown_soft
        - shutdown_hard
        - reboot_soft
        - reboot_hard
        - migrate_online
        - migrate_offline
        - snapshot_withram
        - snapshot_dskonly
        - snapshot_reverse
        - snapshot_remove
        @remark Cette fonction est déjà définie
        @retval array<string> La liste des opérations possibles.
    */
    public function getOperationsList(){
        $operations[] = 'start';
        $operations[] = 'clone';
        $operations[] = 'shutdown_soft';
        $operations[] = 'shutdown_hard';
        $operations[] = 'reboot_soft';
        $operations[] = 'reboot_hard';
        $operations[] = 'migrate_online';
        $operations[] = 'migrate_offline';
        $operations[] = 'snapshot_withram';
        $operations[] = 'snapshot_dskonly';
        $operations[] = 'snapshot_reverse';
        $operations[] = 'snapshot_remove';
        return $operations;
    }

    /**
        @brief Transforme une action + cleanly en type d'operation
        @details
        - start : mode toujours a null
        - shutdown/reboot : mode pour cleanly
        - migrate : mode pour migrate
        - snapshot : mode pour withRAM
        @remark Cette fonction est déjà définie
        @param string $action start, clone, shutdown, ...
        @param boolean|NULL $mode TRUE si on veut cleanly|withram|soft|online
        @retval string Une opération de getOperationsList() ou $action
    */
    public function operationToString($action,$mode=NULL){
        $operation = strtolower($action);
        if( $mode === NULL ){
            return $operation;
        }
        if( is_bool( $mode )){
            if( $operation == 'snapshot'){
                if( $mode ) $operation.= '_withram';
                else        $operation.= '_dskonly';
            }
            elseif( $operation == 'migrate' ){
                if( $mode ) $operation.= '_online';
                else        $operation.= '_offline';
            }
            else{
                if( $mode ) $operation.= '_soft';
                else        $operation.= '_hard';
            }
        }
        elseif($mode != NULL){
            $operation.= '_'.strtolower($mode);
        }
        return $operation;
    }

    /**
        @brief Détecte si une opération est permise
        @details
        - start : mode toujours a null
        - shutdown/reboot : mode pour cleanly
        - migrate : mode pour migrate
        - snapshot : mode pour withRAM
        @remark Cette fonction est déjà définie
        @param string $action start, clone, shutdown, ...
        @param boolean|NULL $mode TRUE si on veut cleanly|withram|soft|online
        @retval boolean TRUE si l'opération est permise
    */
    public function isAllowedOperation($action,$mode=NULL){
        $operations_enum = $this->getOperationsList();

        //Configuration et vérification des paramètres
        $operation = $this->operationToString($action,$mode);
        $op_arr = explode('_',$operation);

        if( !in_array($operation,$operations_enum)){
            $ret['allowed'] = FALSE;
            $ret['cause'] = "Unknown operation type : $operation";
            return $ret;
        }

        //Aucune opération possible si la VM est locked
        if( $this->isLocked()){
            $ret['allowed'] = FALSE;
            $ret['cause'] = 'Locked VM';
            return $ret;
        }

        //start/reboot/shutdown
        if( $operation == 'start'){
            if( $this->isAlive() ){
                $ret['allowed'] = FALSE;
                $ret['cause'] = 'VM already running';
                return $ret;
            }
        }
        elseif( $operation == 'shutdown' || $operation == 'reboot' ){
            $state = $this->getState();
            if( $state == 'stopped' ){
                $ret['allowed'] = FALSE;
                $ret['cause'] = 'VM not running';
                return $ret;
            }
            elseif( $this->getState() == 'paused' && $op_arr[1] == 'soft' ){
                $ret['allowed'] = FALSE;
                $ret['cause'] = 'VM is paused';
                return $ret;
            }
        }
        else if( $operation == 'migrate' || $operation == 'snapshot' ){
            $state = $this->getState();
            if( $state == 'stopped' && in_array($exp[1],array('online','withRAM'))){ //hot_migrate or withRAM_snapshot
                $ret['allowed'] = FALSE;
                $ret['cause'] = 'VM not running';
                return $ret;
            }
        }
        //Si la machine n'est pas locked, une migration est toujours permise

        $ret['allowed'] = TRUE;
        return $ret;
    }

    /**
        @brief Retourne la totalité des opérations connues de l'API permises sur la VM
        @remark Cette fonction est déjà définie
        @retval array<string> La liste des opérations permises.
    */
    public function getAllowedOperations($all=TRUE){
        $operations = $this->getOperationsList();
        foreach($operations as $operation){
            $ret[$operation] = $this->isAllowedOperation($operation)['allowed'];
        }
        return $ret;
    }

    /**
        @brief Permet d'accèder aux donées de l'objet en utilisant des $keys.
        @details
        Permet d'accèder aux données de l'objet sans passer par des méthodes.
        Peut donc être utiliser dans des boucles ou autre choses.
        @remark Cette fonction est déjà définie
        @param string $key C'est le nom de la donnée à récupérer.
        @param NULL|string $format Si $format vaut id ou name, transformera la
            donnée renvoyée en id ou name ou liste d'id/name.
            Par défaut = NULL ne transformera pas la donnée.
    */
    public function getField($key,$format=NULL){
        $key = strtolower($key);
        $ret = NULL;
        if    ( $key == 'state')        $ret = $this->getState();
        elseif( $key == 'is_locked')    $ret = $this->isLocked();
        elseif( $key == 'is_ha')        $ret = $this->isHA();
        elseif( $key == 'ha_settings')  $ret = $this->getHASettings();
        elseif( $key == 'lock_status')  $ret = $this->getLockStatus();
        elseif( $key == 'host')         $ret = $this->getHost();
        elseif( $key == 'cpu_platform') $ret = $this->getCpuPlatform();
        elseif( $key == 'is_agent')     $ret = $this->isAgent();

        elseif( $key == 'volumes')            $ret = $this->getVolumes();
        elseif( $key == 'snapshots')          $ret = $this->getSnapshots();
        elseif( $key == 'allowed_operations') $ret = $this->getAllowedOperations();

        elseif( $key == 'disk_read')    $ret = $this->getDiskRead();
        elseif( $key == 'disk_write')   $ret = $this->getDiskWrite();

        else return parent::getField($key,$format);

        if( $key == 'allowed_operations'){
            $tmp = array();
            foreach($ret as $key => $value){
                if( $value ) $tmp[] = $key;
            }
            return implode(',',$tmp);
        }
        return cluster::formatField($ret,$format);
    }

    /**
        @brief Transforme un objet en un array
        @remark Cette fonction est déjà définie
        @details
        Cette fonction peut être utilisée sans paramètres. Ou on peut préciser
        Si on souhaite récupérer les métriques et les listes.
        Généralement, la récupération de metriques prends du temps à travers
        les APIs hyperviseurs, pareils pour les listes.
        Si vous souhaitez choisir les champs à récupéer, utiliser plutot
        convertObjectToArray()
        - Les champs basiques retournés sont : "id,name,description,type,manufacturer,os,os_version,address,uptime,alive,state,is_locked,lock_status,is_ha,ha_settings,host,cpu_platform,is_agent"
        - Les champs metriques retournés sont : "cpus_usage,cpus_total,mem_usage,mem_total,net_in,net_out,disk_read,disk_write"
        - Les champs listes retournés sont : "volumes,snapshots,allowed_operations"
        @param boolean $enable_metrics Retourne aussi les données de type metrics (plus lent)
        @param boolean $enable_lists Retourne aussi les données de type list (plus lent)
        @param boolean $human FALSE pour des données bruts. TRUE pour des données 'compréhensibles'.
        @retval array Un tableau associatif contenant les clés de $basics et $lists
    */
    public function toArray($enable_metrics=FALSE,$enable_lists=FALSE,$human=FALSE){
        $str = 'id,name,description,type,manufacturer,os,os_version,address,uptime,alive';
        $str.= ',state,is_locked,lock_status,is_ha,ha_settings,host,cpu_platform,is_agent';
        $basics     = explode(',',$str);

        $lists = $from_list = array();
        if( $enable_lists )     $lists  = explode(',','volumes,snapshots,allowed_operations');
        if( $enable_metrics )   $from_list  = explode(',','cpus_usage,cpus_total,mem_usage,mem_total,net_in,net_out,disk_read,disk_write');

        $basics = array_merge($basics,$lists,$from_list);
        return cluster::convertObjectToArray($this,$basics,$lists,$human);
    }

    /**
        @brief Identique a toArray() mais utilisée pour les templates.
          (ils doivent être retournés sous forme d'array et non sous forme d'objects)
        @remark Cette fonction est déjà définie
        @details
        Si vous souhaitez choisir les champs à récupéer, utiliser plutot
        convertObjectToArray()
        - Les champs retournés sont : id,name,isLocked,lock_status,manufacturer,host,os,os_version,volumes,cpus_total,mem_total
    */
    public function toTemplateArray(){
        $enable_metrics=TRUE;
        $enable_lists=TRUE;
        $human=FALSE;
        $str = 'id,name,isLocked,lock_status,manufacturer,host,os,os_version';
        $basics     = explode(',',$str);

        $lists = $from_list = array();
        if( $enable_lists )     $lists  = explode(',','volumes');
        if( $enable_metrics )   $from_list  = explode(',','cpus_total,mem_total');

        $basics = array_merge($basics,$lists,$from_list);
        return cluster::convertObjectToArray($this,$basics,$lists,$human);
    }

    /**
        @brief Implémentation de l'interface JsonSerializable
        @details
        Permet de serialiser en json l'objet. De manière à l'envoyer directement
        à la méthode json_encode. Utilise la fonction toArray() sans paramètres et donc
        ne sérialise que les informatiques basiques (pas de métriques ou list).
    */
    public function jsonSerialize() {
        return $this->toArray();
    }
}

/**
  @brief Représente un stockage. Doit être étendues par les classes des driveurs.
  @details
  L'API d'Esia considére un stockage partagé comme étant un seul stockage connecté
    à plusieurs clusters.
  @remark Certaines méthodes doivent être redéfinies.
*/
class Storage implements JsonSerializable{
    /**
      @brief Le Constructeur. Contient toujours au minimum un ID
      @param string $id Un identifiant permettant identifier le stockage
    */
    public function __construct ($id=NULL){
    }

    /**
      @brief Retourne l'id du stockage
      @retval string $id Un identifiant permettant identifier le stockage
    */
    public function getId(){
        return NULL;
    }

    /**
      @brief Retourne le nom du stockage.
      @warning Le nom du stockage n'est pas forcément unique pour un stockage local.
        Il faut alors ajouter le nom de l'hyperviseur pour avoir une unicité.
      @retval string $id Un identifiant permettant identifier le stockage
    */
    public function getName(){
        return NULL;
    }

    /**
      @brief Retourne la capacité totale du stockage en octets.
      @retval integer La capacité totale du stockage en bytes
    */
    public function getSizeTotal(){
        return NULL;
    }

    /**
      @brief Retourne l'utilisation réelle du stockage.
      @retval integer L'utilisation totale du stockage en bytes
    */
    public function getSizeUsed(){
        return NULL;
    }

    /**
      @brief Retourne l'allocation réelle du stockage en bytes.
      @details
      Cette valeur correspond à la capacité totale de chaque disque virtuel.
      En fonction du format du disque virtuel, l'utilisation réelle est soit, par
      rapport à la taille virtuelle :
      - Plus petite, elle correspond alors à l'espace disque utilisé sur le disque virtuel.
      - Identique, le format du disque virtuel alloue directement 8 GB pour un disque de 8 GB.
      Du coup, l'espace alloué (SizeVirtual), pourrait être plus grande que la
      capacité totale (SizeTotal) du stockage. Excepté si l'ensemble des disques virtuels
      ont un format correspondant à l'usage réel du disque.
      @retval integer L'allocation virtuelle du stockage.
    */
    public function getSizeVirtual(){
        return NULL;
    }

    /**
      @brief Indique si le stockage est partagé ou pas.
      @retval boolean TRUE si le stockage est partagé. FALSE si il est local.
    */
    public function isShared(){
        return NULL;
    }

    /**
      @brief Récupére l'ensemble des hyperviseurs connectés au stockage.
      @details
      XAPI de Xenserver permet de faire la différence entre les hyperviseurs connectés ou pas.
      Proxmox récupére tous les hyperviseurs, je soupçonne sans être sur que c'est uniquement
      les hyperviseurs connectés.
      Pour être certains d'avoir le même résultat dans toute l'API. Il est décidé
      que cette méthode doit faire en sorte de ne retourner que les hyperviseurs
      connectés au stockage. Si Cluster::getHosts() retourne 3 hypevsieurs et que
      Storage::getHosts() en retourne 2, alors il y a un hyperviseur non connecté.
      @retval array<Storage> La liste des hyperviseurs actuellement connectés au stockage.
    */
    public function getHosts(){
        return NULL;
    }

    /**
      @brief Retourne le type de contenu stocké sur ce stockage.
      @details
      Les infos retournés dépendent du système de virtualisation. Par exemple :
      - Pour Proxmox : 'images' ou 'vztmpl,iso,backups'
      - Pour Xenserver : 'user', 'disk', 'iso'
      @retval string Retourne un mot ou une liste de mot séparés par des virgules ','.
    */
    public function getContent(){
        return NULL;
    }

    /**
      @brief Retourne le type de stockage
      @details
      Les infos retournés dépendent du système de virtualisation. Par exemple :
      - Pour Proxmox : 'dir', 'rbd', 'zfspool'
      - Pour XenServer : 'lvm', 'udev'
      @retval string Une indication sur le type du stockage.
    */
    public function getType(){
        return NULL;
    }

    /**
        @brief Retourne une liste de tableau associatifs de volumes.
        @details
        Ci-dessous un exemple de tableau de volumes
        ```php
          $volumes_list = array();
          $volume['id']= $id:
          $volume['name']='Mon volume';
          $volume['desc']='Ceci est mon volume';
          $volume['size_used']= 8589934592; //8 GB
          $volume['size_virt']= 8589934592;
          $volume['parent']= NULL;
          $volume['storage']= $array_storages;
          $volume['vm'] = $array_vm;
          $volume['io_read']= 9127
          $volume['io_write']= 350
          $volumes_list[] = $volume;
          return $volumes_list;
        ```
        Les éléments sont :
        - 'id' : L'id du volume
        - 'name' : Le nom du volume
        - 'desc' : La description du volume
        - 'size_used' : La taille du disque sur le storage.
            Soit inférieure, soit identique à size_virt (allocation dynamique ou pas).
        - 'size_virt' : La taille allouée au storage.
        - 'parent' : L'id du volume parent si il y en a un. NULL sinon.
        - 'storage' : Une tableau contenant des objets Storage.
        - 'vm' : Un tableau contenant des objets VMs (C'est rare mais il peut y en avoir plusieurs)
        - 'io_read' : Vitesses de lecture sur ce volume en Bytes/s
        - 'io_write' : Vitesses d'écritue sur ce volume en Bytes/s
        @retval array<array> La liste de tableaux associatifs décrit ci-dessus.
    */
    public function getVolumes(){
        return NULL;
    }

    /**
        @brief Retourne les vitesses de lecture disque (ensemble des disques de la VM) en Bytes/s
        @retval integer Bytes/s
    */
    function getDiskRead(){
        return NULL;
    }

    /**
        @brief Retourne les vitesses d'écriture disque (ensemble des disques de la VM) en Bytes/s
        @retval integer Bytes/s
    */
    function getDiskWrite(){
        return NULL;
    }

      /**
        @brief Permet d'accèder aux donées de l'objet en utilisant des $keys.
        @details
        Permet d'accèder aux données de l'objet sans passer par des méthodes.
        Peut donc être utiliser dans des boucles ou autre choses.
        @remark Cette fonction est déjà définie
        @param string $key C'est le nom de la donnée à récupérer.
        @param NULL|string $format Si $format vaut id ou name, transformera la
            donnée renvoyée en id ou name ou liste d'id/name.
            Par défaut = NULL ne transformera pas la donnée.
    */
    public function getField($key,$format=NULL){
        $key = strtolower($key);
        $ret = NULL;
        if    ( $key == 'id')       $ret = $this->getID();
        elseif( $key == 'name')     $ret = $this->getName();
        elseif( $key == 'is_shared')$ret = $this->isShared();
        elseif( $key == 'content')  $ret = $this->getContent();
        elseif( $key == 'type')     $ret = $this->getType();

        elseif( $key == 'volumes')  $ret = $this->getVolumes();
        elseif( $key == 'hosts')    $ret = $this->getHosts();

        elseif( $key == 'disk_read')    $ret = $this->getDiskRead();
        elseif( $key == 'disk_write')   $ret = $this->getDiskWrite();
        elseif( $key == 'size_usage')   $ret = $this->getSizeUsed();
        elseif( $key == 'size_virtual') $ret = $this->getSizeVirtual();
        elseif( $key == 'size_total')   $ret = $this->getSizeTotal();

        return cluster::formatField($ret,$format);
    }

    /**
          @brief Transforme un objet en un array
          @remark Cette fonction est déjà définie
          @details
          Cette fonction peut être utilisée sans paramètres. Ou on peut préciser
          Si on souhaite récupérer les métriques et les listes.
          Généralement, la récupération de metriques prends du temps à travers
          les APIs hyperviseurs, pareils pour les listes.
          Si vous souhaitez choisir les champs à récupéer, utiliser plutot
          convertObjectToArray()
          - Les champs basiques retournés sont : "id,name,is_shared,content,type"
          - Les champs metriques retournés sont : "disk_read,disk_write,size_usage,size_virtual,size_total"
          - Les champs listes retournés sont : "hosts,volumes"
          @param boolean $enable_metrics Retourne aussi les données de type metrics (plus lent)
          @param boolean $enable_lists Retourne aussi les données de type list (plus lent)
          @param boolean $human FALSE pour des données bruts. TRUE pour des données 'compréhensibles'.
          @retval array Un tableau associatif contenant les clés de $basics et $lists
      */
    public function toArray($enable_metrics=FALSE,$enable_lists=FALSE,$human=FALSE){
        $basics     = explode(',','id,name,is_shared,content,type');

        $lists = $from_list = array();
        if( $enable_lists )     $lists  = explode(',','hosts,volumes');
        if( $enable_metrics )   $from_list  = explode(',','disk_read,disk_write,size_usage,size_virtual,size_total');

        $basics = array_merge($basics,$lists,$from_list);
        return cluster::convertObjectToArray($this,$basics,$lists,$human);
    }

    /**
        @brief Implémentation de l'interface JsonSerializable
        @details
        Permet de serialiser en json l'objet. De manière à l'envoyer directement
        à la méthode json_encode. Utilise la fonction toArray() sans paramètres et donc
        ne sérialise que les informatiques basiques (pas de métriques ou list).
    */
    public function jsonSerialize() {
        return $this->toArray();
    }
}
