<?php
include_once "XapiClient.php";
include_once "XenRRDClient.php";
include_once "common/common.class.php";

/**
    @file
    @brief Fichier drivers pour les clusters XenServer
    @details
    Ce fichier contient un ensemble de classes qui étendent les classes du
    fichier common.class.php pour l'hyperviseur XenServer.

    Ce fichier inclut les fichiers XapiClient qui gére les requêtes envoyés à
    un hyperviseur XenServer.
    Les classes dans ce fichier manipulent la classe XenXapiClient de ce fichier
    et formattent ces données en respectant les conventions décrient dans common.class.php

    Un second fichier XenRRDClient est inclut et contient la classe XenRRDClient.
    Cette dernière fait la même chose que XenXapiClient mais est utilisée pour
    récupérer des données RRDs. Les RRDs passant par une autre API que les autres données.
    @author Julien George
*/

class XenServerCluster extends Cluster{
    protected $driver;
    protected $rrd_driver;
    protected $cluster_status=NULL;

    protected $ref_list = array();
    protected $ref_byname = array();
    protected $item_list = array();

    protected $rrd_cache = array();

    public function __construct($settings){
        $this->driver = new XenXapiClient($settings);
        $this->rrd_driver = new XenRRDClient($settings);

        try{
            $this->driver->open();
        }
        catch(SessionException $ex){
            if( $ex->getType() == 'HOST_IS_SLAVE'){
                $settings['host'] = $ex->getErrorDescription()[1];
                $this->driver = new XenXapiClient($settings);
                $this->driver->open();
            }
            else throw $ex;
        }
        $this->rrd_driver->open();
    }

    public function cleanCache(){
        $ref_list = array();
        $item_list = array();

        $rrd_cache = array();
    }

    public function getDriver(){
        return $this->driver;
    }

    public function getRRDValues($addr){
        if(!isset($this->rrd_cache[$addr])){
            $this->rrd_driver->setAddress($addr);
            $this->rrd_cache[$addr] = $this->rrd_driver->get_rrd(TRUE,120);
        }
        return $this->rrd_cache[$addr];
    }

    private function load_data($type){
        $data = $this->driver->get("$type.get_all_records");

        foreach($data as $ref => $row){
            $this->ref_list[$type][$row['uuid']]=$ref;
            if( isset($row['name_label']) ){ //Beaucoups d'objects xenserver ont un name_label mais pas tous
                if( !isset($this->ref_byname[$type][$row['name_label']]) ) $this->ref_byname[$type][$row['name_label']]=array($ref);
                else $this->ref_byname[$type][$row['name_label']][] = $ref;
            }
            $this->item_list[$type][$ref] = $row;
        }
    }

    public function getRefList($type){
        if( !isset($this->ref_list[$type])) $this->load_data($type);
        return $this->ref_list[$type];
    }

    public function getRefByUuid($type,$uuid){
        if( !isset($this->ref_list[$type])) $this->load_data($type);
        if( !isset($this->ref_list[$type][$uuid])) throw new Exception("$type with uuid $uuid not found \n");
        return $this->ref_list[$type][$uuid];
    }

    public function getRawByRef($type,$ref){
        if( $ref == 'OpaqueRef:NULL' ) return NULL;
        if( !isset($this->item_list[$type])) $this->load_data($type);
        if( !isset($this->item_list[$type][$ref])) throw new Exception("$type with ref $ref not found \n");
        return $this->item_list[$type][$ref];
    }

    public function getRawByUuid($type,$uuid){
        $ref = $this->getRefByUuid($type,$uuid);
        return $this->getRawByRef($type,$ref);
    }

    public function getRefListByName($type,$name){
        if( !isset($this->ref_byname[$type])) $this->load_data($type);
        if( !isset($this->ref_byname[$type][$name])) return array();
        return $this->ref_byname[$type][$name];
    }

    protected function loadClusterData(){
        if( $this->cluster_status == NULL ){
            $this->cluster_status = array_fill_keys(array('name','master','is_cluster','is_ha'),NULL);

            $this->load_data('pool');
            $data = $this->item_list['pool'];

            if( empty($data) ){
                $this->load_data('host');
                $data = $this->item_list['host'];

                $row = array_pop($data);
                $this->cluster_status['name']		= $row['name_label'];
                $this->cluster_status['master']      = $row['uuid'];
                $this->cluster_status['is_cluster']  = FALSE;
                $this->cluster_status['is_ha']		= FALSE;
            }
            else{
                $row = array_pop($data);
                $this->cluster_status['name'] 		= $row['name_label'];
                $this->cluster_status['master'] 		= $this->driver->get('host.get_record',$row['master'])['uuid'];
                $this->cluster_status['is_cluster'] 	= !in_array('cluster_create',$row['allowed_operations']);
                $this->cluster_status['is_ha']		= ($row['ha_enabled'] == 1);
            }
        }
    }

    public function jsonSerialize() {
        $this->loadClusterData();
        return $this->cluster_status;
    }

    public function getName():string{
        $this->loadClusterData();
        return $this->cluster_status['name'];
    }
    public function getMasterHost():string{
        $this->loadClusterData();
        return $this->cluster_status['master'];
    }

    public function isCluster():bool{
        $this->loadClusterData();
        return $this->cluster_status['is_cluster'];
    }

    public function isHA():bool{
        $this->loadClusterData();
        return $this->cluster_status['is_ha'];
    }

    public function getHosts():array{
        $host_list_uuid = array_keys($this->getRefList('host'));
        foreach($host_list_uuid as $uuid){
            $host_list[] = new XenServerHost($uuid,$this);
        }
        return $host_list;
    }
    public function getVMs():array{
        $vm_list_uuid = array_keys($this->getRefList('VM'));
        foreach($vm_list_uuid as $uuid){
            $raw_data =  $this->getRawByUuid('VM',$uuid);
            if($raw_data['is_a_template'] == 0 && $raw_data['is_a_snapshot'] != 1){
                $vm_list[] = new XenServerVM($uuid,$this);
            }
        }
        return $vm_list;
    }

    public function getTemplates():array{
        $vm_list_uuid = array_keys($this->getRefList('VM'));
        foreach($vm_list_uuid as $uuid){
            $raw_data =  $this->getRawByUuid('VM',$uuid);
            if($raw_data['is_a_template'] == 1  && $raw_data['is_a_snapshot'] != 1 ){
                $vm = new XenServerVM($uuid,$this);
                $tpl = $vm->toTemplateArray();
                if( empty($tpl['os']) ){
                    if( preg_match('/(\D+)(.*)/', $tpl['name'], $matches) ){
                        if( isset($matches[1])) $tpl['os'] = $matches[1];
                        if( isset($matches[2])) $tpl['os_version'] = $matches[2];
                    }
                }
                $vm_list[] =$tpl;
            }
        }
        return $vm_list;
    }

    public function getStorages($shared=NULL):array{
      $sr_list_uuid = array_keys($this->getRefList('SR'));
      foreach($sr_list_uuid as $uuid){
          $sr_list[] = new XenServerStorage($uuid,$this);
      }
      if( $shared === NULL || !is_bool($shared) ){
          return $sr_list;
      }
      else{
          $ret = array();
          foreach($sr_list as $storage){
              if( $storage->isShared() == $shared ){
                  $ret[] = $storage;
              }
          }
          return $ret;
      }
    }

    public function getVMFromSnapshot($snapid){
        $snap_data =  $this->getRawByUuid('VM',$snapid);
        $vm_data = $this->getRawByRef('VM',$snap_data['snapshot_of']);
        return $this->getVM($vm_data['uuid']);
    }

    function getVMByName($name){
        $ret = array();
        foreach($this->getRefListByName('VM',$name) as $ref){
            $raw_data =  $this->getRawByRef('VM',$ref);
            if($raw_data['is_a_template'] == 0 && $raw_data['is_a_snapshot'] != 1){
                $ret[] = new XenServerVM($raw_data['uuid'],$this);
            }
        }
        return $ret;
    }

    function getHostByName($name){
        $ref = $this->getRefListByName('host',$name);
        $raw_data =  $this->getRawByRef('host',array_shift($ref));
        return new XenServerHost($raw_data['uuid'],$this);
    }

    public function getVM($vmid,$refresh=FALSE){
        if($refresh){
            //On recharge le cache des VMs
            $this->load_data('VM');
        }
        $raw_data =  $this->getRawByUuid('VM',$vmid);
        if( $raw_data == NULL ) return NULL;

        return new XenServerVM($vmid,$this);
    }

    public function cloneVM($source,$target_name,$full=TRUE){
        //Analyse de la source
        $vmid=NULL;
        $host=NULL;
        if( is_string($source)){ //VMID
            $vmid = $source;
        }
        elseif(is_array($source)&& isset($source['id'])){ //TEMPLATE
            $vmid=$source['id'];
        }
        elseif(is_object($source)){ //Object
            $vmid = $source->getId();
        }

        if( $vmid == NULL ){
            throw new Exception("Unable to found the VM/template source\n");
        }

        $ref = $this->getRefByUuid('VM',$vmid);

        //Start
        try{
            if($full){
                $data = $this->driver->get('Async.VM.copy',array($ref,$target_name,NULL));
            }
            else{
                $data = $this->driver->get('Async.VM.clone',array($ref,$target_name));
            }
        }
        catch(Exception $ex){
            throw $ex;
        }

        //On efface le cache des VMs
        unset($this->ref_list['VM']);
        unset($this->item_list['VM']);
    }
}

class XenServerHost extends Host{
    protected $cluster;
    protected $ref;

    public function __construct($id,XenServerCluster $cluster){
        $this->cluster = $cluster;

        $this->ref = $cluster->getRefByUuid('host',$id);
        $raw_data = $cluster->getRawByRef('host',$this->ref);

        $this->setId($id);
        $this->setName($raw_data['name_label']);
        $this->setDescription('name_description');
        $this->setType('hypervisor');
        $this->setManufacturer('Citrix');
        $this->setOS('XenServer');
        $this->setOSVersion($raw_data['software_version']['product_version']);
        $this->setAddress($raw_data['address']);
    }

    public function getCluster(){
        return $this->cluster;
    }

    public function getUpTime():int{
        $raw_data = $this->cluster->getRawByRef('host',$this->ref);
        if( !isset($raw_data['other_config']['boot_time']) ) return 0;
        return time() - $raw_data['other_config']['boot_time'];
    }

    public function getCpusTotal():int{
        $raw_data = $this->cluster->getRawByRef('host',$this->ref);
        if( !isset($raw_data['other_config']['boot_time']) ) return 0;
        return $raw_data['cpu_info']['cpu_count'];
    }

    public function getCpusUsage():float{
        $rrds = $this->cluster->getRRDValues($this->getAddress());
        return $rrds['host'][$this->getId()]['cpu_avg'];
    }

    public function getMemTotal():int{
        $rrds = $this->cluster->getRRDValues($this->getAddress());
        return $rrds['host'][$this->getId()]['memory_total_kib']*1024;
    }

    public function getMemUsage():int{
        $rrds = $this->cluster->getRRDValues($this->getAddress());
        $rrd = $rrds['host'][$this->getId()];
        return round($rrd['memory_total_kib']*1024-$rrd['memory_free_kib']*1024);
    }

    public function getNetInUsage():int{
        $rrds = $this->cluster->getRRDValues($this->getAddress());
        return $rrds['host'][$this->getId()]['pif_aggr_rx'];
    }

    public function getNetOutUsage():int{
        $rrds = $this->cluster->getRRDValues($this->getAddress());
        return $rrds['host'][$this->getId()]['pif_aggr_tx'];
    }

    public function getNetInterfaces(){
      $rrds = $this->cluster->getRRDValues($this->getAddress());
      $pif_speed = array();
      $pif_speed['global']['rx'] = $this->getNetInUsage();
      $pif_speed['global']['tx'] = $this->getNetOutUsage();
      foreach($rrds['host'][$this->getId()] as $key => $value)
      {
        $type = 0;
        $ifname = 1;
        $dir = 2;

        $exp = explode('_',$key);
        if($exp[$type] == 'pif' && $exp[$ifname] != 'aggr')
        {
          $pif_speed[$exp[$ifname]][$exp[$dir]] = round($value);
        }
      }

      $raw_data = $this->cluster->getRawByRef('host',$this->ref);
      $pifs_refs = $raw_data['PIFs'];

      $ifaces = array();
      $ifaces['global'] = $pif_speed['global'];
      foreach($pifs_refs as $pif_ref)
      {
          $pif_raw_data = $this->cluster->getRawByRef('PIF',$pif_ref);
          $retRow['device'] = $pif_raw_data['device'];
          $addresses = array();
          foreach(array_merge(array($pif_raw_data['IP']),$pif_raw_data['IPv6']) as $addr)
          {
            if( !empty($addr)) $addresses[] = $addr;
          }

          $retRow['ip'] = implode(',',$addresses);
          $retRow['mac'] = $pif_raw_data['MAC'];
          $retRow['mtu'] = $pif_raw_data['MTU'];
          $retRow['rx'] = $pif_speed[$pif_raw_data['device']]['rx'];
          $retRow['tx'] = $pif_speed[$pif_raw_data['device']]['tx'];
          $retRow['speed'] = NULL;

          if( !empty($pif_raw_data['metrics']) )
          {
            $metrics = $this->cluster->getRawByRef('PIF_metrics',$pif_raw_data['metrics']);
            $retRow['speed'] = $metrics['speed'];
          }

          $ifaces[$pif_raw_data['device']] = $retRow;
      }
      return $ifaces;
    }

    public function isAlive():bool{
        $resource_data = $this->cluster->getRawByRef('host',$this->ref);
        $metrics_data = $this->cluster->getRawByRef('host_metrics',$resource_data['metrics']);
        return ( $resource_data['enabled'] && $metrics_data['live'] );
    }

    public function getVMs():array{
        $vm_list = array();
        foreach($this->cluster->getVMs() as $vm){
            if( $vm->getHost() != NULL && $vm->getHost()->getId() == $this->getId()){
                $vm_list[] = $vm;
            }
        }
        return $vm_list;
    }

    public function getTemplates(){
        $vm_list = array();
        foreach($this->cluster->getTemplates() as $tpl){
            if( $tpl['host'] != NULL && $tpl['host'] == $this->getId()){
                $vm_list[] = $tpl;
            }
        }
        return $vm_list;
    }

    protected function getHostRRD($key){
        $rrds = $this->cluster->getRRDValues($this->getAddress());
        //echo "$key\n";
        //print_r($rrds['host']);
        if( !isset($rrds['host'][$this->getId()][$key]) ) return NULL;
        return $rrds['host'][$this->getId()][$key];
    }

    public function getPBDMetric($name,$sr_uuid){
        $short_uuid = explode('-',$sr_uuid)[0];
        return $this->getHostRRD($name.'_'.$short_uuid);
    }
}

class XenServerVM extends VM{
    protected $cluster;
    protected $host = NULL;
    protected $ref;

    private static $enum_power_states = array(
        'started' => 'running',
        'running' => 'running',
        'stopped' => 'stopped',
        'halted'  => 'stopped',
        'paused'  => 'paused',
        'suspended'  => 'suspended',
    );

    public function __construct($id,XenServerCluster $cluster){
        $this->cluster = $cluster;

        $this->ref = $cluster->getRefByUuid('VM',$id);
        $raw_data = $cluster->getRawByRef('VM',$this->ref);
        $guest_data = $cluster->getRawByRef('VM_guest_metrics',$raw_data['guest_metrics']);

        $this->setId($id);
        $this->setName($raw_data['name_label']);
        $this->setDescription('name_description');
        $this->setType('Virtual server');
        $this->setManufacturer('Citrix');
    }

    private function getVMRRDs($regex){
        if( $this->getHost() == NULL ) return NULL;

        $rrds = $this->cluster->getRRDValues($this->getHost()->getAddress());
        $vm_rrds = $rrds['vm'][$this->getId()];
        $rrd_list = array();
        foreach($vm_rrds as $name => $row){
            if( preg_match($regex,$name) ){
                $rrd_list[]=$row;
            }
        }
        return $rrd_list;
    }

    public function getHost(){
        if( $this->host === NULL){
            $raw_data = $this->cluster->getRawByRef('VM',$this->ref);

            $host_ref = $raw_data['resident_on'];
            if( $host_ref == 'OpaqueRef:NULL' ){
                $host_ref = $raw_data['affinity'];
            }
            if( $host_ref == 'OpaqueRef:NULL' ){
                $this->host = FALSE;
            }
            else{
                $host_data = $this->cluster->getRawByRef('host',$raw_data['resident_on']);
                $this->host = new XenServerHost($host_data['uuid'],$this->cluster);
            }
        }

        if($this->host == FALSE){
            return NULL;
        }
        else{
            return $this->host;
        }
    }

    public function getAddress(){
        if( parent::getAddress() == NULL){
            $raw_data = $this->cluster->getRawByRef('VM',$this->ref);
            $networks = $this->cluster->getRawByRef('VM_guest_metrics',$raw_data['guest_metrics'])['networks'];
            if( $networks == NULL ){
                $this->setAddress(NULL);
            }
            else{
                ksort($networks);
                $first_net = array_shift($networks);
                $first_addr = explode(' ',$first_net)[0];
                $this->setAddress($first_addr);
            }
        }
        return parent::getAddress();
    }

    public function getOS(){
        if( parent::getOS() == NULL ){
            $raw_data = $this->cluster->getRawByRef('VM',$this->ref);
            $os = $this->cluster->getRawByRef('VM_guest_metrics',$raw_data['guest_metrics'])['os_version'];
            $this->setOS($os['distro']);
        }
        return parent::getOS();
    }

    public function getOSVersion(){
        if( parent::getOSVersion() == NULL ){
            $raw_data = $this->cluster->getRawByRef('VM',$this->ref);
            $os = $this->cluster->getRawByRef('VM_guest_metrics',$raw_data['guest_metrics'])['os_version'];
            $this->setOSVersion($os['uname']);
        }
        return parent::getOSVersion();
    }

    public function getState(){
        $raw_data = $this->cluster->getRawByRef('VM',$this->ref);
        return self::$enum_power_states[strtolower($raw_data['power_state'])];
    }

    public function isLocked($refresh=FALSE){
        $status = $this->getLockStatus($refresh);
        return ( $status != NULL );
    }

    public function getLockStatus($refresh=FALSE){
        if( $refresh ){
            return $this->cluster->getVM($this->getID(),TRUE)->getLockStatus();
        }
        $raw_data = $this->cluster->getRawByRef('VM',$this->ref);
        if( empty($raw_data['current_operations']) ) return NULL;
        return $raw_data['current_operations'];
    }

    public function isAlive(){
        return $this->getState() == 'running' ;
    }

    public function getUpTime(){
        if( !$this->isAlive()) return NULL;
        $raw_data = $this->cluster->getRawByRef('VM',$this->ref);
        $start_time = $this->cluster->getRawByRef('VM_metrics',$raw_data['metrics'])['start_time'];
        if( is_object($start_time)) return $start_time->timestamp;
        elseif(isset($start_time['timestamp'])) return $start_time['timestamp'];
        else return NULL;
    }

    public function getCpuPlatform(){
        $raw_data = $this->cluster->getRawByRef('VM',$this->ref);
        return json_encode($raw_data['platform']);
    }

    public function isAgent(){
        return NULL;
    }

    public function getCpusTotal(){
        $raw_data = $this->cluster->getRawByRef('VM',$this->ref);
        return $raw_data['VCPUs_max'];
    }

    public function getCpusUsage(){
        $rrd_list = $this->getVMRRDs('/^cpu\d?/');
        if( empty($rrd_list)) return NULL;

        return array_sum($rrd_list);
    }

    public function getMemTotal(){
        $raw_data = $this->cluster->getRawByRef('VM',$this->ref);
        return $raw_data['memory_dynamic_max'];
    }

    public function getMemUsage(){
        $rrd_list = $this->getVMRRDs('/^memory_internal_free$/');
        if( empty($rrd_list)) return NULL;

        $value = array_sum($rrd_list)*1024;
        return round($this->getMemTotal() - $value,2);
    }

    public function getNetInUsage(){
        $rrd_list = $this->getVMRRDs('/^vif_\d+_rx/');
        if( empty($rrd_list)) return NULL;

        return array_sum($rrd_list);
    }

    public function getNetOutUsage(){
        $rrd_list = $this->getVMRRDs('/^vif_\d+_tx/');
        if( empty($rrd_list)) return NULL;

        return array_sum($rrd_list);
    }

    public function getDiskRead(){
        $rrd_list = $this->getVMRRDs('/^vbd_.*_read/');
        if( empty($rrd_list)) return NULL;

        return array_sum($rrd_list);
    }

    public function getDiskWrite(){
        $rrd_list = $this->getVMRRDs('/^vbd_.*_write/');
        if( empty($rrd_list)) return NULL;

        return array_sum($rrd_list);
    }

    public function getHASettings(){
        $raw_data = $this->cluster->getRawByRef('VM',$this->ref);
        return $raw_data['ha_restart_priority'];
    }

    public function isHA(){
        return ($this->getHASettings() == 'restart' && $this->cluster->isHA());
    }

    public function getSnapshots(){
        $raw_data = $this->cluster->getRawByRef('VM',$this->ref);
        $snapshots = array();
        foreach($raw_data['snapshots'] as $snap_ref){
            $snapshots[] = $this->cluster->getRawByRef('VM',$snap_ref);
        }

        $pve_keys['name']='name_label';
        $pve_keys['description']='name_description';
        $pve_keys['parent']='parent';
        $pve_keys['time']='snapshot_time';
        $pve_keys['withRAM']='vmstate';

        $ret = array();
        foreach($snapshots as $row){
            $ret_row = array();
            $ret_row['id'] = $row['uuid'];
            foreach($pve_keys as $ekey => $pkey ){
                  if( isset($row[$pkey] ) )  $ret_row[$ekey] = $row[$pkey];
                  else $ret_row[$ekey] = NULL;
            }
            $ret_row['withRAM'] = NULL;
            $ret_row['time'] = $ret_row['time']->timestamp;
            if( isset( $resource[$ret_row['parent']])) {
                  $ret_row['parent'] = $resource[$ret_row['parent']]['uuid'];
            }
            else{
                  $ret_row['parent'] = NULL;
            }
            $ret[]=$ret_row;
        }
        return $ret;
    }

    public function getVolumes(){
        $ret = array();
        $raw_data = $this->cluster->getRawByRef('VM',$this->ref);
        foreach($raw_data['VBDs'] as $vbd_ref){
            $raw_vbd = $this->cluster->getRawByRef('VBD',$vbd_ref);
            $raw_vdi = $this->cluster->getRawByRef('VDI',$raw_vbd['VDI']);

            if($raw_vdi == NULL) continue;

            $retRow = array();
            $retRow['id'] = $raw_vdi['uuid'];
            $retRow['name'] = $raw_vdi['name_label'];
            $retRow['desc'] = $raw_vdi['name_description'];
            $retRow['size_used'] = $raw_vdi['physical_utilisation'];
            $retRow['size_virt'] = $raw_vdi['virtual_size'];
            $parent = NULL;
            if( $raw_vdi['parent'] != NULL ){
                $parent = $this->cluster->getRawByRef('VDI',$raw_vdi['parent']);
            }
            if( $parent != NULL ) $parent = $parent['uuid'];
            $retRow['parent'] = $parent;
            $retRow['storage'] = $this;
            $retRow['io_read'] = 0;
            $retRow['io_write'] = 0;


            foreach($raw_vdi['VBDs'] as $vbd_ref){
                $vbd_raw = $this->cluster->getRawByRef('VBD',$vbd_ref);

                $device = $raw_vbd['device'];
                $rrd_read = $this->getVMRRDs('/^vbd_'.$device.'_read/');
                $rrd_write = $this->getVMRRDs('/^vbd_'.$device.'_write/');

                $retRow['io_read'] += array_shift($rrd_read);
                $retRow['io_write']+= array_shift($rrd_write);

                $vm_raw = $this->cluster->getRawByRef('VM',$vbd_raw['VM']);
                if( $vm_raw != NULL ) {
                    if( $vm_raw['uuid'] == $this->getID()){
                        $vm_list[] = $this;
                    }
                    else{
                        $vm_list[] = new XenServerVM($vm_raw['uuid'],$this->cluster);
                    }
                }
            }
            $retRow['vm']= $vm_list;
            $ret[]=$retRow;
        }
        return $ret;
    }

    public function start(){
        $allowed = $this->isAllowedOperation('start');
        if( !$allowed['allowed'] ) throw new Exception($allowed['cause']);

        $state = $this->getState();
        $vmid = $this->getId();
        $ref = $this->cluster->getRefByUuid('VM',$vmid);

        if( $state == 'suspended' ){
            $data = $this->cluster->getDriver()->get('Async.VM.resume',array($ref,FALSE,FALSE));
        }
        elseif( $state == 'paused' ){
            $data = $this->cluster->getDriver()->get('Async.VM.unpause',array($ref));
        }
        else{
            $data = $this->cluster->getDriver()->get('Async.VM.start',array($ref,FALSE,FALSE));
        }
    }

    public function shutdown($cleanly=TRUE){
        $allowed = $this->isAllowedOperation('shutdown',$cleanly);
        if( !$allowed['allowed'] ) throw new Exception($allowed['cause']);

        $vmid = $this->getId();
        $ref = $this->cluster->getRefByUuid('VM',$vmid);

        if( !$cleanly ){
            $data = $this->cluster->getDriver()->get('Async.VM.hard_shutdown',array($ref));
        }
        else{
            $data = $this->cluster->getDriver()->get('Async.VM.clean_shutdown',array($ref));
        }
    }

    public function reboot($cleanly=TRUE){
        $allowed = $this->isAllowedOperation('reboot',$cleanly);
        if( !$allowed['allowed'] ) throw new Exception($allowed['cause']);

        $vmid = $this->getId();
        $ref = $this->cluster->getRefByUuid('VM',$vmid);

        if( !$cleanly ){
            $data = $this->cluster->getDriver()->get('Async.VM.hard_reboot',array($ref));
        }
        else{
            $data = $this->cluster->getDriver()->get('Async.VM.clean_reboot',array($ref));
        }
    }

    public function isAllowedOperation($key,$mode=NULL){
        $operation = $this->operationToString($key,$mode);
        $op_arr = explode('_',$operation);

        if($op_arr[0] == 'migrate' ){
            $ret['allowed'] = FALSE;
            $ret['cause'] = 'Operation not available for XenServer driver';
            return $ret;
        }
        if($operation == 'snapshot_withram' ){
            $ret['allowed'] = FALSE;
            $ret['cause'] = 'Operation not available for XenServer driver';
            return $ret;
        }
        return parent::isAllowedOperation($operation,NULL);
    }

    public function migrate($target,$online){
        throw new Exception('Not implemented');

        $allowed = $this->isAllowedOperation('migrate',$online);
        if( !$allowed['allowed'] ) throw new Exception($allowed['cause']);
        //Not implemented for Xenserver Driver. Is available into XAPI API but not documented.
    }

    public function cloneVM($target_name,$full=TRUE){
        $allowed = $this->isAllowedOperation('clone');
        if( !$allowed['allowed'] ) throw new Exception($allowed['cause']);
        $this->getCluster()->cloneVM($this,$target_name,TRUE);
    }

    public function createSnapshot($name,$description='',$withRam=FALSE){
        $allowed = $this->isAllowedOperation('snapshot',$withRam);
        if( !$allowed['allowed'] ) throw new Exception($allowed['cause']);

        $vmid = $this->getId();
        $ref = $this->cluster->getRefByUuid('VM',$vmid);

        $refsnap=$this->cluster->getDriver()->get('VM.snapshot',array($ref,$name));
        $this->cluster->getDriver()->get('VM.set_name_description',array($refsnap,$description));
    }

    public function reverseToSnapshot($snapid){
        $allowed = $this->isAllowedOperation('snapshot_reverse');
        if( !$allowed['allowed'] ) throw new Exception($allowed['cause']);

        $ref = $this->cluster->getRefByUuid('VM',$snapid);
        $this->cluster->getDriver()->get('VM.revert',array($ref));
    }

    public function removeSnapshot($snapid){
        $allowed = $this->isAllowedOperation('snapshot_remove');
        if( !$allowed['allowed'] ) throw new Exception($allowed['cause']);

        $ref = $this->cluster->getRefByUuid('VM',$snapid);
        $this->cluster->getDriver()->get('VM.destroy',array($ref));
    }
}

class XenServerStorage extends Storage{
    protected $cluster;
    protected $ref;

    public function __construct($id,XenServerCluster $cluster){
        $this->cluster = $cluster;
        $this->ref = $cluster->getRefByUuid('SR',$id);
    }

    public function getId(){
        $raw_data = $this->cluster->getRawByRef('SR',$this->ref);
        return $raw_data['uuid'];
    }

    public function getName(){
      $raw_data = $this->cluster->getRawByRef('SR',$this->ref);
      return $raw_data['name_label'];
    }

    public function getSizeTotal(){
      $raw_data = $this->cluster->getRawByRef('SR',$this->ref);
      $value = $raw_data['physical_size'];
      if( $value == -1 ) $value = 0;
      return $value;
    }

    public function getSizeUsed(){
      $raw_data = $this->cluster->getRawByRef('SR',$this->ref);
      $value = $raw_data['physical_utilisation'];
      if( $value == -1 ) $value = 0;
      return $value;
    }

    public function getSizeVirtual(){
        $raw_data = $this->cluster->getRawByRef('SR',$this->ref);
        $value = $raw_data['virtual_allocation'];
        if( $value == -1 ) $value = 0;
        return $value;
    }

    public function isShared(){
      $raw_data = $this->cluster->getRawByRef('SR',$this->ref);
      return $raw_data['shared'] == 1;
    }

    public function getContent(){
      $raw_data = $this->cluster->getRawByRef('SR',$this->ref);
      return $raw_data['content_type'];
    }

    public function getType(){
        $raw_data = $this->cluster->getRawByRef('SR',$this->ref);
        return $raw_data['type'];
    }

    public function getHosts($connected=TRUE){
        $raw_data = $this->cluster->getRawByRef('SR',$this->ref);
        $host = array();
        foreach($raw_data['PBDs'] as $pbd_ref){
            $pbd = $this->cluster->getRawByRef('PBD',$pbd_ref);
            $is_connected = ($pbd['currently_attached'] == 1);
            if( $connected == 'all' || $is_connected == $connected ){
              $host_uuid = $this->cluster->getRawByRef('host',$pbd['host'])['uuid'];
              $host[] = new XenServerHost($host_uuid,$this->cluster);
          }
        }
        return $host;
    }

    public function getVolumes(){
        $ret = array();
        $raw_data = $this->cluster->getRawByRef('SR',$this->ref);
        foreach($raw_data['VDIs'] as $vdi_ref){
            $raw_vdi = $this->cluster->getRawByRef('VDI',$vdi_ref);
            $retRow['id'] = $raw_vdi['uuid'];
            $retRow['name'] = $raw_vdi['name_label'];
            $retRow['desc'] = $raw_vdi['name_description'];
            $retRow['size_used'] = $raw_vdi['physical_utilisation'];
            $retRow['size_virt'] = $raw_vdi['virtual_size'];

            $parent = $this->cluster->getRawByRef('VDI',$raw_vdi['parent']);
            if( $parent != NULL ) $parent = $parent['uuid'];
            $retRow['parent'] = $parent;
            $retRow['storage'] = $this;
            $retRow['io_read'] = 0;
            $retRow['io_write'] = 0;

            $vm_list = array();
            foreach($raw_vdi['VBDs'] as $vbd_ref){
                $vbd_raw = $this->cluster->getRawByRef('VBD',$vbd_ref);

                $vm_raw = $this->cluster->getRawByRef('VM',$vbd_raw['VM']);
                if( $vm_raw != NULL ) {
                    $vm = new XenServerVM($vm_raw['uuid'],$this->cluster);
                    if( $vm->isAlive()){
                        foreach($vm->getVolumes() as $vl){
                            if( $vl['id'] == $retRow['id']){
                                $retRow['io_read'] += $vl['io_read'];
                                $retRow['io_write'] += $vl['io_write'];
                            }
                        }
                    }
                    $vm_list[] = $vm;
                }
            }
            $retRow['vm'] =$vm_list;
            $ret[] = $retRow;
        }
        return $ret;
    }

    public function getDiskRead(){
        $host_list = $this->getHosts(TRUE);
        $total = 0;
        foreach($host_list as $nd){
            $total += $nd->getPBDMetric('read',$this->getID());
        }
        return $total;
    }

    public function getDiskWrite(){
        $host_list = $this->getHosts(TRUE);
        $total = 0;
        foreach($host_list as $nd){
            $total += $nd->getPBDMetric('write',$this->getID());
        }
        return $total;
    }
}
